/*
* Copyright (c) OpenSim project, http://sim.opensecondlife.org/
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the <organization> nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY <copyright holder> ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL <copyright holder> BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
* 
*/


using System;
using System.Collections.Generic;
using System.Net;
using System.Net.Sockets;
using System.IO;
using libsecondlife;
using OpenSim;

namespace OpenSim.GridServers
{
	/// <summary>
	/// Handles connection to Grid Servers.
	/// also Sim to Sim connections?
	/// </summary>
	
	
	
	
	public interface IGridServer
	{
		bool RequestConnection();
		Dictionary<uint, agentcircuitdata> agentcircuits {
			get;
			set;
		}
		UUIDBlock RequestUUIDBlock();
		void RequestNeighbours(); //should return a array of neighbouring regions
		AuthenticateResponse AuthenticateSession(LLUUID sessionID, LLUUID agentID, uint circuitCode);
		bool LogoutSession(LLUUID sessionID, LLUUID agentID, uint circuitCode);
		void SetServerInfo(string ServerUrl, string ServerKey);
	 	void AddNewSession(Login session); // only used by local version of grid server 
	 									   // and didn't use to be part of this interface until we put this in a dll
	}
	
	public struct UUIDBlock
	{
		public LLUUID BlockStart;
		public LLUUID BlockEnd;
	}
	
	public class AuthenticateResponse
	{
		public bool Authorised;
		public Login LoginInfo;
		
		public AuthenticateResponse()
		{
			
		}
			
	}
	
	public class Login
    {
    	public string First = "Test";
    	public string Last = "User";
    	public LLUUID Agent;
    	public LLUUID Session;
    	public LLUUID InventoryFolder;
    	public LLUUID BaseFolder;
    	public Login()
    	{
    		
    	}
    }

	public interface IGridPlugin
	{
		IGridServer GetGridServer();
	}

       public class agentcircuitdata {
                public agentcircuitdata() { }
                public LLUUID AgentID;
                public LLUUID SessionID;
                public LLUUID SecureSessionID;
                public string firstname;
                public string lastname;
                public uint circuitcode;
        }


}
